#include <stdio.h>

typedef struct _nodo;

typedef struct _funcion {
	struct _funcion * sig;
	char nombre[20];
	struct _nodo * cuerpo;
	unsigned short numeroParametros;
} Funcion;

typedef struct _nodo {
	struct _nodo * izq;
	struct _nodo * dch;
	unsigned short tipo;
	int valor;
	Funcion * ptrFuncion;
} Nodo;

typedef struct _id {
	struct _id * sig;
	char nombre[20];
	short posicion;
} Id;

typedef struct _pila {
	struct _pila * sig;
	int valor;
} NodoPila;

#define TINDEFINIDO 0
#define TCASO 	    1
#define TCOND		 2
#define TAND		 3
#define TOR		    4
#define TNOT		 5
#define TMAYORQUE  6
#define TMENORQUE  7
#define TIGUALQUE  8
#define TMAS		 9
#define TMENOS	    10
#define TPOR		 11
#define TDIVIDIDO  12
#define TID		    13
#define TNUM		 14
#define TFUNCION	 15
#define TEXPR		 16

Nodo * crearNodo(){
	Nodo * x;
	x = (Nodo *)malloc(sizeof(Nodo));
	x->izq = x-> dch = NULL;
	x->ptrFuncion = NULL;
	return x;
}

NodoPila * crearNodoPila(){
	NodoPila * x;
	x = (NodoPila *)malloc(sizeof(NodoPila));
	x->sig = NULL;
	return x;
}

Id * crearId(){
	Id * x;
	x = (Id *)malloc(sizeof(Id));
	x->sig = NULL;
	x->posicion = 0;
	return x;
}

Funcion * crearFuncion(){
	Funcion * x;
	x = (Funcion *)malloc(sizeof(Funcion));
	x->sig = NULL;
	x->cuerpo = NULL;
	x->numeroParametros = 0;
	return x;
}

void insertarFuncion(Funcion ** f, Funcion * ptr){
	ptr->sig = (*f);
	(*f) = ptr;
}

void insertarPila(NodoPila ** f, NodoPila * ptr){
	ptr->sig = (*f);
	(*f) = ptr;
}

void insertarId(Id ** t, Id * ptr){
	ptr->sig = (*t);
	(*t) = ptr;
}

void crearTablaId(Id ** t){
	(*t) = NULL;
}

void destruirTablaId(Id ** t){
	while ((*t) != NULL){
		Id * ptr = (*t);
		(*t) = (*t)->sig;
		free(ptr);
	}
}

void borrarPila(NodoPila ** p){
	while ((*p) != NULL){
		NodoPila * ptr = (*p);
		(*p) = (*p)->sig;
		free(ptr);
	}
}

Id * buscarId(Id * t, char * nombre){
	while ((t != NULL)&&(strcmp(t->nombre, nombre))){
		t = t->sig;
	}
	return t;
}

Funcion * buscarFuncion(Funcion * t, char * nombre){
	while ((t != NULL)&&(strcmp(t->nombre, nombre))){
		t = t->sig;
	}
	return t;
}


/* Aqui viene la funcion principal, que se encarga de evaluar un cuerpo */

int evaluarCuerpo(Nodo * x, NodoPila ** p){
#define pila ((*p))
	/*NodoPila * a=pila;
	int k=0;
	printf("%d -", x->tipo);
	while (a!=NULL){
		printf("%d-", a->valor);
		a=a->sig;k++; //if (k>5) exit(0);
	} printf("\n");*/
	switch (x->tipo) {
		case (TINDEFINIDO) : { return 0;}
		case (TNUM)		: { return x->valor;}
		case (TID)		: { 
									NodoPila * ptr = pila;
									int i;
									for (i = x->valor; i > 1; i--)
										ptr = ptr->sig;
									return ptr->valor;
								}
		case (TFUNCION):	{
									int resultado, i=0;
									Nodo * ptr;
									/* Evaluar parametros reales y meterlos en una pila auxiliar*/
									NodoPila * pilaAux=NULL;
									NodoPila * nuevoNodoPila;
									for (ptr = x->izq; ptr->tipo == TEXPR; ptr = ptr->dch) {
										nuevoNodoPila = crearNodoPila();
										nuevoNodoPila->valor = evaluarCuerpo(ptr->izq, &pila);
										insertarPila(&pilaAux, nuevoNodoPila);
									}
									nuevoNodoPila = crearNodoPila();
									nuevoNodoPila->valor = evaluarCuerpo(ptr, &pila);
									insertarPila(&pilaAux, nuevoNodoPila);
									/* Volcar resultados en la pila principal */
									while (pilaAux != NULL){
										NodoPila * ptr = pilaAux;
										pilaAux = pilaAux->sig;
										insertarPila(&pila, ptr);
									}
									/* Invocar la funcion */
									resultado = evaluarCuerpo(x->ptrFuncion->cuerpo, &pila);
									/* Desapilar */
									for (i=x->ptrFuncion->numeroParametros; i>0; i--){
										NodoPila * ptr = pila;
										pila = pila->sig;
										free(ptr);
									}
									return resultado;
								}
		case (TCASO)	:	{
									while ((x->tipo == TCASO) && (!evaluarCuerpo(x->izq->izq, &pila))){
										x = x->dch;
									}
									if (x->tipo == TCASO) {
										return evaluarCuerpo(x->izq->dch, &pila);
									} else {
										return evaluarCuerpo(x, &pila);
									}
								}
		case (TMAS)		:	{ return evaluarCuerpo(x->izq, &pila) + evaluarCuerpo(x->dch, &pila); }
		case (TPOR)		:	{ return evaluarCuerpo(x->izq, &pila) * evaluarCuerpo(x->dch, &pila); }
		case (TMENOS)	:	{ return evaluarCuerpo(x->izq, &pila) - evaluarCuerpo(x->dch, &pila); }
		case (TDIVIDIDO):	{ return evaluarCuerpo(x->izq, &pila) / evaluarCuerpo(x->dch, &pila); }
		case (TAND)		:	{ return evaluarCuerpo(x->izq, &pila) && evaluarCuerpo(x->dch, &pila); }
		case (TOR)		:	{ return evaluarCuerpo(x->izq, &pila) || evaluarCuerpo(x->dch, &pila); }
		case (TNOT)		:	{ return !evaluarCuerpo(x->izq, &pila); }
		case (TMAYORQUE)	:	{ return evaluarCuerpo(x->izq, &pila) > evaluarCuerpo(x->dch, &pila); }
		case (TMENORQUE)	:	{ return evaluarCuerpo(x->izq, &pila) < evaluarCuerpo(x->dch, &pila); }
		case (TIGUALQUE)	:	{ return evaluarCuerpo(x->izq, &pila) == evaluarCuerpo(x->dch, &pila); }
		default			:	{ return 0; }
	}
#undef pila
}
